@@ -1,8 +1,9 @@
-from typing import Union, Tuple, List
+from typing import Optional, Union, Tuple, List
 
 import os
 import csv
 import gym
+import gymnasium
 import operator
 import datetime
 import functools
@@ -13,23 +14,24 @@ from torch.utils.data.sampler import BatchSampler
 
 
 class Memory:
-    def __init__(self, 
-                 memory_size: int, 
-                 num_envs: int = 1, 
-                 device: Union[str, torch.device] = "cuda:0", 
-                 export: bool = False, 
-                 export_format: str = "pt", 
+    def __init__(self,
+                 memory_size: int,
+                 num_envs: int = 1,
+                 device: Optional[Union[str, torch.device]] = None,
+                 export: bool = False,
+                 export_format: str = "pt",
                  export_directory: str = "") -> None:
         """Base class representing a memory with circular buffers
 
         Buffers are torch tensors with shape (memory size, number of environments, data size).
         Circular buffers are implemented with two integers: a memory index and an environment index
-        
+
         :param memory_size: Maximum number of elements in the first dimension of each internal storage
         :type memory_size: int
         :param num_envs: Number of parallel environments (default: 1)
         :type num_envs: int, optional
-        :param device: Device on which a torch tensor is or will be allocated (default: "cuda:0")
+        :param device: Device on which a torch tensor is or will be allocated (default: ``None``).
+                       If None, the device will be either ``"cuda:0"`` if available or ``"cpu"``
         :type device: str or torch.device, optional
         :param export: Export the memory to a file (default: False).
                        If True, the memory will be exported when the memory is filled
@@ -45,7 +47,7 @@ class Memory:
         """
         self.memory_size = memory_size
         self.num_envs = num_envs
-        self.device = torch.device(device)
+        self.device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu") if device is None else torch.device(device)
 
         # internal variables
         self.filled = False
@@ -54,6 +56,10 @@ class Memory:
 
         self.tensors = {}
         self.tensors_view = {}
+        self.tensors_keep_dimensions = {}
+
+        self.sampling_indexes = None
+        self.all_sequence_indexes = np.concatenate([np.arange(i, memory_size * num_envs + i, num_envs) for i in range(num_envs)])
 
         # exporting data
         self.export = export
@@ -65,7 +71,7 @@ class Memory:
 
     def __len__(self) -> int:
         """Compute and return the current (valid) size of the memory
-        
+
         The valid size is calculated as the ``memory_size * num_envs`` if the memory is full (filled).
         Otherwise, the ``memory_index * num_envs + env_index`` is returned
 
@@ -73,28 +79,43 @@ class Memory:
         :rtype: int
         """
         return self.memory_size * self.num_envs if self.filled else self.memory_index * self.num_envs + self.env_index
-        
-    def _get_space_size(self, space: Union[int, Tuple[int], gym.Space]) -> int:
+
+    def _get_space_size(self,
+                        space: Union[int, Tuple[int], gym.Space, gymnasium.Space],
+                        keep_dimensions: bool = False) -> Union[Tuple, int]:
         """Get the size (number of elements) of a space
 
         :param space: Space or shape from which to obtain the number of elements
-        :type space: int, tuple or list of integers, or gym.Space
+        :type space: int, tuple or list of integers, gym.Space, or gymnasium.Space
+        :param keep_dimensions: Whether or not to keep the space dimensions (default: False)
+        :type keep_dimensions: bool
 
         :raises ValueError: If the space is not supported
 
-        :return: Size of the space data
-        :rtype: Space size (number of elements)
+        :return: Size of the space. If keep_dimensions is True, the space size will be a tuple
+        :rtype: int or tuple of int
         """
         if type(space) in [int, float]:
-            return int(space)
+            return (int(space),) if keep_dimensions else int(space)
         elif type(space) in [tuple, list]:
-            return np.prod(space)
+            return tuple(space) if keep_dimensions else np.prod(space)
         elif issubclass(type(space), gym.Space):
             if issubclass(type(space), gym.spaces.Discrete):
-                return 1
+                return (1,) if keep_dimensions else 1
             elif issubclass(type(space), gym.spaces.Box):
-                return np.prod(space.shape)
+                return tuple(space.shape) if keep_dimensions else np.prod(space.shape)
             elif issubclass(type(space), gym.spaces.Dict):
+                if keep_dimensions:
+                    raise ValueError("keep_dimensions=True cannot be used with Dict spaces")
+                return sum([self._get_space_size(space.spaces[key]) for key in space.spaces])
+        elif issubclass(type(space), gymnasium.Space):
+            if issubclass(type(space), gymnasium.spaces.Discrete):
+                return (1,) if keep_dimensions else 1
+            elif issubclass(type(space), gymnasium.spaces.Box):
+                return tuple(space.shape) if keep_dimensions else np.prod(space.shape)
+            elif issubclass(type(space), gymnasium.spaces.Dict):
+                if keep_dimensions:
+                    raise ValueError("keep_dimensions=True cannot be used with Dict spaces")
                 return sum([self._get_space_size(space.spaces[key]) for key in space.spaces])
         raise ValueError("Space type {} not supported".format(type(space)))
 
@@ -142,28 +163,34 @@ class Memory:
         with torch.no_grad():
             self.tensors[name].copy_(tensor)
 
-    def create_tensor(self, name: str, size: Union[int, Tuple[int], gym.Space], dtype: Union[torch.dtype, None] = None) -> bool:
+    def create_tensor(self,
+                      name: str,
+                      size: Union[int, Tuple[int], gym.Space, gymnasium.Space],
+                      dtype: Optional[torch.dtype] = None,
+                      keep_dimensions: bool = False) -> bool:
         """Create a new internal tensor in memory
-        
+
         The tensor will have a 3-components shape (memory size, number of environments, size).
         The internal representation will use _tensor_<name> as the name of the class property
 
         :param name: Tensor name (the name has to follow the python PEP 8 style)
         :type name: str
         :param size: Number of elements in the last dimension (effective data size).
-                     The product of the elements will be computed for collections or gym spaces types
-        :type size: int, tuple or list of integers or gym.Space
+                     The product of the elements will be computed for sequences or gym/gymnasium spaces
+        :type size: int, tuple or list of integers, gym.Space, or gymnasium.Space
         :param dtype: Data type (torch.dtype).
                       If None, the global default torch data type will be used (default)
         :type dtype: torch.dtype or None, optional
-        
+        :param keep_dimensions: Whether or not to keep the dimensions defined through the size parameter (default: False)
+        :type keep_dimensions: bool
+
         :raises ValueError: The tensor name exists already but the size or dtype are different
 
         :return: True if the tensor was created, otherwise False
         :rtype: bool
         """
         # compute data size
-        size = self._get_space_size(size)
+        size = self._get_space_size(size, keep_dimensions)
         # check dtype and size if the tensor exists
         if name in self.tensors:
             tensor = self.tensors[name]
@@ -172,10 +199,15 @@ class Memory:
             if dtype is not None and tensor.dtype != dtype:
                 raise ValueError("The dtype of the tensor {} ({}) doesn't match the existing one ({})".format(name, dtype, tensor.dtype))
             return False
+        # define tensor shape
+        tensor_shape = (self.memory_size, self.num_envs, *size) if keep_dimensions else (self.memory_size, self.num_envs, size)
+        view_shape = (-1, *size) if keep_dimensions else (-1, size)
         # create tensor (_tensor_<name>) and add it to the internal storage
-        setattr(self, "_tensor_{}".format(name), torch.zeros((self.memory_size, self.num_envs, size), device=self.device, dtype=dtype))
+        setattr(self, "_tensor_{}".format(name), torch.zeros(tensor_shape, device=self.device, dtype=dtype))
+        # update internal variables
         self.tensors[name] = getattr(self, "_tensor_{}".format(name))
-        self.tensors_view[name] = self.tensors[name].view(-1, self.tensors[name].size(-1))
+        self.tensors_view[name] = self.tensors[name].view(*view_shape)
+        self.tensors_keep_dimensions[name] = keep_dimensions
         # fill the tensors (float tensors) with NaN
         for tensor in self.tensors.values():
             if torch.is_floating_point(tensor):
@@ -210,7 +242,7 @@ class Memory:
 
         - number of environments less than num_envs:
           Store the samples and increment the environment index (second index) by the number of the environments
-        
+
         - number of environments equals num_envs:
           Store the samples and increment the memory index (first index) by one
 
@@ -224,21 +256,21 @@ class Memory:
             raise ValueError("No samples to be recorded in memory. Pass samples as key-value arguments (where key is the tensor name)")
 
         # dimensions and shapes of the tensors (assume all tensors have the dimensions of the first tensor)
-        tmp = tensors[next(iter(tensors))]
+        tmp = tensors.get("states", tensors[next(iter(tensors))])  # ask for states first
         dim, shape = tmp.ndim, tmp.shape
 
-        # multi environment (number of environments less than num_envs)
-        if dim == 2 and shape[0] < self.num_envs:
-            for name, tensor in tensors.items():
-                if name in self.tensors:
-                    self.tensors[name][self.memory_index, self.env_index:self.env_index + tensor.shape[0]].copy_(tensor)
-            self.env_index += tensor.shape[0]
         # multi environment (number of environments equals num_envs)
-        elif dim == 2 and shape[0] == self.num_envs:
+        if dim == 2 and shape[0] == self.num_envs:
             for name, tensor in tensors.items():
                 if name in self.tensors:
                     self.tensors[name][self.memory_index].copy_(tensor)
             self.memory_index += 1
+        # multi environment (number of environments less than num_envs)
+        elif dim == 2 and shape[0] < self.num_envs:
+            for name, tensor in tensors.items():
+                if name in self.tensors:
+                    self.tensors[name][self.memory_index, self.env_index:self.env_index + tensor.shape[0]].copy_(tensor)
+            self.env_index += tensor.shape[0]
         # single environment - multi sample (number of environments greater than num_envs (num_envs = 1))
         elif dim == 2 and self.num_envs == 1:
             for name, tensor in tensors.items():
@@ -273,7 +305,11 @@ class Memory:
             if self.export:
                 self.save(directory=self.export_directory, format=self.export_format)
 
-    def sample(self, names: Tuple[str], batch_size: int, mini_batches: int = 1) -> List[List[torch.Tensor]]:
+    def sample(self,
+               names: Tuple[str],
+               batch_size: int,
+               mini_batches: int = 1,
+               sequence_length: int = 1) -> List[List[torch.Tensor]]:
         """Data sampling method to be implemented by the inheriting classes
 
         :param names: Tensors names from which to obtain the samples
@@ -282,9 +318,11 @@ class Memory:
         :type batch_size: int
         :param mini_batches: Number of mini-batches to sample (default: 1)
         :type mini_batches: int, optional
-        
+        :param sequence_length: Length of each sequence (default: 1)
+        :type sequence_length: int, optional
+
         :raises NotImplementedError: The method has not been implemented
-        
+
         :return: Sampled data from tensors sorted according to their position in the list of names.
                  The sampled tensors will have the following shape: (batch size, data size)
         :rtype: list of torch.Tensor list
@@ -310,29 +348,47 @@ class Memory:
             return [[self.tensors_view[name][batch] for name in names] for batch in batches]
         return [[self.tensors_view[name][indexes] for name in names]]
 
-    def sample_all(self, names: Tuple[str], mini_batches: int = 1) -> List[List[torch.Tensor]]:
+    def sample_all(self, names: Tuple[str], mini_batches: int = 1, sequence_length: int = 1) -> List[List[torch.Tensor]]:
         """Sample all data from memory
-        
+
         :param names: Tensors names from which to obtain the samples
         :type names: tuple or list of strings
         :param mini_batches: Number of mini-batches to sample (default: 1)
         :type mini_batches: int, optional
+        :param sequence_length: Length of each sequence (default: 1)
+        :type sequence_length: int, optional
 
         :return: Sampled data from memory.
                  The sampled tensors will have the following shape: (memory size * number of environments, data size)
         :rtype: list of torch.Tensor list
         """
+        # sequential order
+        if sequence_length > 1:
+            if mini_batches > 1:
+                batches = BatchSampler(self.all_sequence_indexes, batch_size=len(self.all_sequence_indexes) // mini_batches, drop_last=True)
+                return [[self.tensors_view[name][batch] for name in names] for batch in batches]
+            return [[self.tensors_view[name][self.all_sequence_indexes] for name in names]]
+
+        # default order
         if mini_batches > 1:
             indexes = np.arange(self.memory_size * self.num_envs)
             batches = BatchSampler(indexes, batch_size=len(indexes) // mini_batches, drop_last=True)
             return [[self.tensors_view[name][batch] for name in names] for batch in batches]
         return [[self.tensors_view[name] for name in names]]
-        
+
+    def get_sampling_indexes(self) -> Union[tuple, np.ndarray, torch.Tensor]:
+        """Get the last indexes used for sampling
+
+        :return: Last sampling indexes
+        :rtype: tuple or list, numpy.ndarray or torch.Tensor
+        """
+        return self.sampling_indexes
+
     def save(self, directory: str = "", format: str = "pt") -> None:
         """Save the memory to a file
 
         Supported formats:
-        
+
         - PyTorch (pt)
         - NumPy (npz)
         - Comma-separated values (csv)
@@ -350,7 +406,7 @@ class Memory:
         os.makedirs(os.path.join(directory, "memories"), exist_ok=True)
         memory_path = os.path.join(directory, "memories", \
             "{}_memory_{}.{}".format(datetime.datetime.now().strftime("%y-%m-%d_%H-%M-%S-%f"), hex(id(self)), format))
-        
+
         # torch
         if format == "pt":
             torch.save({name: self.tensors[name] for name in self.get_tensor_names()}, memory_path)
@@ -359,7 +415,7 @@ class Memory:
             np.savez(memory_path, **{name: self.tensors[name].cpu().numpy() for name in self.get_tensor_names()})
         # comma-separated values
         elif format == "csv":
-            # open csv writer
+            # open csv writer # TODO: support keeping the dimensions
             with open(memory_path, "a") as file:
                 writer = csv.writer(file)
                 names = self.get_tensor_names()
@@ -391,18 +447,18 @@ class Memory:
             data = torch.load(path)
             for name in self.get_tensor_names():
                 setattr(self, "_tensor_{}".format(name), data[name])
-        
+
         # numpy
         elif path.endswith(".npz"):
             data = np.load(path)
             for name in data:
                 setattr(self, "_tensor_{}".format(name), torch.tensor(data[name]))
-        
+
         # comma-separated values
         elif path.endswith(".csv"):
             # TODO: load the memory from a csv
             pass
-        
+
         # unsupported format
         else:
             raise ValueError("Unsupported format: {}".format(path))
