@@ -1,6 +1,7 @@
-from typing import Union, Tuple
+from typing import Optional, Union, Tuple
 
 import gym
+import gymnasium
 import numpy as np
 
 import torch
@@ -8,47 +9,52 @@ import torch.nn as nn
 
 
 class RunningStandardScaler(nn.Module):
-    def __init__(self, 
-                 size: Union[int, Tuple[int], gym.Space], 
-                 epsilon: float = 1e-8, 
+    def __init__(self,
+                 size: Union[int, Tuple[int], gym.Space, gymnasium.Space],
+                 epsilon: float = 1e-8,
                  clip_threshold: float = 5.0,
-                 device: Union[str, torch.device] = "cuda:0") -> None:
+                 device: Optional[Union[str, torch.device]] = None) -> None:
         """Standardize the input data by removing the mean and scaling by the standard deviation
 
-        The implementation is adapted from the rl_games library 
+        The implementation is adapted from the rl_games library
         (https://github.com/Denys88/rl_games/blob/master/rl_games/algos_torch/running_mean_std.py)
 
         Example::
 
             >>> running_standard_scaler = RunningStandardScaler(size=2)
-            >>> data = ...  # tensor of shape (N, 2)
+            >>> data = torch.rand(3, 2)  # tensor of shape (N, 2)
             >>> running_standard_scaler(data)
+            tensor([[0.1954, 0.3356],
+                    [0.9719, 0.4163],
+                    [0.8540, 0.1982]])
 
         :param size: Size of the input space
-        :type size: int, tuple or list of integers, or gym.Space
-        :param epsilon: Small number to avoid division by zero (default: 1e-8)
+        :type size: int, tuple or list of integers, gym.Space, or gymnasium.Space
+        :param epsilon: Small number to avoid division by zero (default: ``1e-8``)
         :type epsilon: float
-        :param clip_threshold: Threshold to clip the data (default: 5.0)
+        :param clip_threshold: Threshold to clip the data (default: ``5.0``)
         :type clip_threshold: float
-        :param device: Device on which a torch tensor is or will be allocated (default: "cuda:0")
+        :param device: Device on which a torch tensor is or will be allocated (default: ``None``).
+                       If None, the device will be either ``"cuda:0"`` if available or ``"cpu"``
         :type device: str or torch.device, optional
         """
         super().__init__()
 
         self.epsilon = epsilon
         self.clip_threshold = clip_threshold
+        self.device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu") if device is None else torch.device(device)
 
         size = self._get_space_size(size)
-        
-        self.register_buffer("running_mean", torch.zeros(size, dtype = torch.float64, device=device))
-        self.register_buffer("running_variance", torch.ones(size, dtype = torch.float64, device=device))
-        self.register_buffer("current_count", torch.ones((), dtype = torch.float64, device=device))
 
-    def _get_space_size(self, space: Union[int, Tuple[int], gym.Space]) -> int:
+        self.register_buffer("running_mean", torch.zeros(size, dtype = torch.float64, device=self.device))
+        self.register_buffer("running_variance", torch.ones(size, dtype = torch.float64, device=self.device))
+        self.register_buffer("current_count", torch.ones((), dtype = torch.float64, device=self.device))
+
+    def _get_space_size(self, space: Union[int, Tuple[int], gym.Space, gymnasium.Space]) -> int:
         """Get the size (number of elements) of a space
 
         :param space: Space or shape from which to obtain the number of elements
-        :type space: int, tuple or list of integers, or gym.Space
+        :type space: int, tuple or list of integers, gym.Space, or gymnasium.Space
 
         :raises ValueError: If the space is not supported
 
@@ -66,10 +72,17 @@ class RunningStandardScaler(nn.Module):
                 return np.prod(space.shape)
             elif issubclass(type(space), gym.spaces.Dict):
                 return sum([self._get_space_size(space.spaces[key]) for key in space.spaces])
+        elif issubclass(type(space), gymnasium.Space):
+            if issubclass(type(space), gymnasium.spaces.Discrete):
+                return 1
+            elif issubclass(type(space), gymnasium.spaces.Box):
+                return np.prod(space.shape)
+            elif issubclass(type(space), gymnasium.spaces.Dict):
+                return sum([self._get_space_size(space.spaces[key]) for key in space.spaces])
         raise ValueError("Space type {} not supported".format(type(space)))
 
     def _parallel_variance(self, input_mean: torch.Tensor, input_var: torch.Tensor, input_count: int) -> None:
-        """Update internal variables using the parallel algorithm for computing variance 
+        """Update internal variables using the parallel algorithm for computing variance
 
         https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance#Parallel_algorithm
 
@@ -95,9 +108,9 @@ class RunningStandardScaler(nn.Module):
 
         :param x: Input tensor
         :type x: torch.Tensor
-        :param train: Whether to train the standardizer (default: False)
+        :param train: Whether to train the standardizer (default: ``False``)
         :type train: bool, optional
-        :param inverse: Whether to inverse the standardizer to scale back the data (default: False)
+        :param inverse: Whether to inverse the standardizer to scale back the data (default: ``False``)
         :type inverse: bool, optional
         """
         if train:
@@ -112,7 +125,7 @@ class RunningStandardScaler(nn.Module):
                 * torch.clamp(x, min=-self.clip_threshold, max=self.clip_threshold) + self.running_mean.float()
         # standardization by centering and scaling
         else:
-            return torch.clamp((x - self.running_mean.float()) / (torch.sqrt(self.running_variance.float()) + self.epsilon), 
+            return torch.clamp((x - self.running_mean.float()) / (torch.sqrt(self.running_variance.float()) + self.epsilon),
                                 min=-self.clip_threshold, max=self.clip_threshold)
 
     def forward(self, x: torch.Tensor, train: bool = False, inverse: bool = False, no_grad: bool = True) -> torch.Tensor:
@@ -125,7 +138,7 @@ class RunningStandardScaler(nn.Module):
             tensor([[0.6933, 0.1905],
                     [0.3806, 0.3162],
                     [0.1140, 0.0272]], device='cuda:0')
-            
+
             >>> running_standard_scaler(x, train=True)
             tensor([[ 0.8681, -0.6731],
                     [ 0.0560, -0.3684],
@@ -138,15 +151,15 @@ class RunningStandardScaler(nn.Module):
 
         :param x: Input tensor
         :type x: torch.Tensor
-        :param train: Whether to train the standardizer (default: False)
+        :param train: Whether to train the standardizer (default: ``False``)
         :type train: bool, optional
-        :param inverse: Whether to inverse the standardizer to scale back the data (default: False)
+        :param inverse: Whether to inverse the standardizer to scale back the data (default: ``False``)
         :type inverse: bool, optional
-        :param no_grad: Whether to disable the gradient computation (default: True)
+        :param no_grad: Whether to disable the gradient computation (default: ``True``)
         :type no_grad: bool, optional
         """
         if no_grad:
             with torch.no_grad():
                 return self._compute(x, train, inverse)
         else:
-            return self._compute(x, train, inverse)
\ No newline at end of file
+            return self._compute(x, train, inverse)
